@@ -42,30 +42,29 @@ You cannot use this program to actually change your UID.\n";
 const char args_doc[] = "NEWROOT CMD [ARG...]";
 
 const struct argp_option options[] = {
-   { "bind",        'b', "SRC[:DST]", 0,
-     "mount SRC at guest DST (default /mnt/0, /mnt/1, etc.)"},
-   { "cd",          'c', "DIR",  0, "initial working directory in container"},
-   { "ch-ssh",       -8, 0,      0, "bind ch-ssh into image"},
-   { "gid",         'g', "GID",  0, "run as GID within container" },
-   { "join",        'j', 0,      0, "use same container as peer ch-run" },
-   { "join-pid",     -5, "PID",  0, "join a namespace using a PID" },
-   { "join-ct",      -3, "N",    0, "number of ch-run peers (implies --join)" },
-   { "join-tag",     -4, "TAG",  0, "label for peer group (implies --join)" },
-   { "no-home",      -2, 0,      0, "don't bind-mount your home directory"},
-   { "no-passwd",    -9, 0,      0, "don't bind-mount /etc/{passwd,group}"},
-   { "private-tmp", 't', 0,      0, "use container-private /tmp" },
-   { "set-env",      -6, "FILE", 0, "set environment variables in FILE"},
-   { "uid",         'u', "UID",  0, "run as UID within container" },
-   { "unset-env",    -7, "GLOB", 0, "unset environment variable(s)" },
-   { "verbose",     'v', 0,      0, "be more verbose (debug if repeated)" },
-   { "version",     'V', 0,      0, "print version and exit" },
-   { "write",       'w', 0,      0, "mount image read-write"},
+   { "bind",          'b', "SRC[:DST]", 0,
+     "mount SRC at guest DST (default: same as SRC)"},
+   { "cd",            'c', "DIR",  0, "initial working directory in container"},
+   { "ch-ssh",         -8, 0,      0, "bind ch-ssh into image"},
+   { "env-no-expand", -10, 0,      0, "don't expand $ in --set-env input"},
+   { "gid",           'g', "GID",  0, "run as GID within container" },
+   { "join",          'j', 0,      0, "use same container as peer ch-run" },
+   { "join-pid",       -5, "PID",  0, "join a namespace using a PID" },
+   { "join-ct",        -3, "N",    0, "number of join peers (implies --join)" },
+   { "join-tag",       -4, "TAG",  0, "label for peer group (implies --join)" },
+   { "no-home",        -2, 0,      0, "don't bind-mount your home directory"},
+   { "no-passwd",      -9, 0,      0, "don't bind-mount /etc/{passwd,group}"},
+   { "private-tmp",   't', 0,      0, "use container-private /tmp" },
+   { "set-env",        -6, "FILE", 0, "set environment variables in FILE"},
+   { "uid",           'u', "UID",  0, "run as UID within container" },
+   { "unset-env",      -7, "GLOB", 0, "unset environment variable(s)" },
+   { "verbose",       'v', 0,      0, "be more verbose (debug if repeated)" },
+   { "version",       'V', 0,      0, "print version and exit" },
+   { "write",         'w', 0,      0, "mount image read-write"},
    { 0 }
 };
 
-/* One possible future here is that fix_environment() ends up in ch_base.c and
-   we add other actions such as SET, APPEND_PATH, etc. */
-enum env_action { END = 0, SET_FILE, UNSET_GLOB };
+enum env_action { ENV_END = 0, ENV_SET, ENV_UNSET };
 
 struct env_delta {
    enum env_action action;
@@ -81,6 +80,8 @@ struct args {
 /** Function prototypes **/
 
 void env_delta_append(struct env_delta **ds, enum env_action act, char *arg);
+void envs_set(char **lines, const int line_ct, const char *filename,
+              const bool expand);
 void fix_environment(struct args *args);
 bool get_first_env(char **array, char **name, char **value);
 int join_ct(int cli_ct);
@@ -112,6 +113,7 @@ int main(int argc, char *argv[])
    args = (struct args){ .c = (struct container){ .ch_ssh = false,
                                                   .container_gid = getegid(),
                                                   .container_uid = geteuid(),
+                                                  .env_expand = true,
                                                   .newroot = NULL,
                                                   .join = false,
                                                   .join_ct = 0,
@@ -176,14 +178,75 @@ void env_delta_append(struct env_delta **ds, enum env_action act, char *arg)
 {
    int i;
 
-   for (i = 0; (*ds)[i].action != END; i++) // count existing
+   for (i = 0; (*ds)[i].action != ENV_END; i++) // count existing
          ;
    T_ (*ds = realloc(*ds, (i+2) * sizeof(struct env_delta)));
-   (*ds)[i+1].action = END;
+   (*ds)[i+1].action = ENV_END;
    (*ds)[i].action = act;
    (*ds)[i].arg = arg;
 }
 
+/* Set environment variables as specified in the array lines, which has length
+   line_ct. filename is the source filename, or NULL if source was not a file;
+   in that case, omit line number from any error messages. If expand, then
+   expand variable notation as described in the man page. */
+void envs_set(char **lines, const int line_ct, const char *filename,
+              const bool expand)
+{
+   char *name, *value_old, *value_new, *lineno_str, *item;
+
+   for (int i = 0; i < line_ct; i++) {
+      bool first_written;
+
+      // Skip blank lines.
+      if (lines[i][0] == 0 || lines[i][0] == '\n')
+         continue;
+
+      // Split line into variable name and value.
+      split(&name, &value_old, lines[i], '=');
+      if (filename == NULL)
+         lineno_str = "";
+      else
+         T_ (1 <= asprintf(&lineno_str, ":%d", i+1));
+      Te (name != NULL, "--set-env: no delimiter: %s%s", filename, lineno_str);
+      Te (name[0] != 0, "--set-env: empty name: %s%s", filename, lineno_str);
+
+      // Strip leading and trailing single quotes from value, if both present.
+      if (   strlen(value_old) >= 2
+          && value_old[0] == '\''
+          && value_old[strlen(value_old) - 1] == '\'') {
+         value_old[strlen(value_old) - 1] = 0;
+         value_old++;
+      }
+
+      // Walk through value fragments separated by colon and expand variables
+      // per documentation.
+      value_new = "";
+      first_written = false;
+      while (1) {                                  // loop executes ≥ once
+         item = strsep(&value_old, ":");           // NULL -> no more items
+         if (item == NULL)
+            break;
+         if (   expand                             // expansion requested
+             && item[0] == '$' && item[1] != 0) {  // ≥1 char in variable name
+            item = getenv(++item);                 // NULL if unset
+            if (item != NULL && item[0] == 0)
+               item = NULL;                        // convert empty to unset
+         }
+         if (item != NULL) {                       // NULL -> omit from output
+            if (first_written)
+               value_new = cat(value_new, ":");
+            value_new = cat(value_new, item);
+            first_written = true;
+         }
+      }
+
+      // Save results.
+      INFO("environment: %s=%s", name, value_new);
+      Z_ (setenv(name, value_new, 1));
+   }
+}
+
 /* Adjust environment variables. */
 void fix_environment(struct args *args)
 {
@@ -212,40 +275,38 @@ void fix_environment(struct args *args)
    }
 
    // --set-env and --unset-env.
-   for (int i = 0; args->env_deltas[i].action != END; i++) {
+   for (int i = 0; args->env_deltas[i].action != ENV_END; i++) {
       char *arg = args->env_deltas[i].arg;
-      if (args->env_deltas[i].action == SET_FILE) {
-         FILE *fp;
-         Tf (fp = fopen(arg, "r"), "--set-env: can't open: %s", arg);
-         for (int j = 1; true; j++) {
-            char *line = NULL;
-            size_t len = 0;
-            errno = 0;
-            if (-1 == getline(&line, &len, fp)) {
-               if (errno == 0)  // EOF
-                  break;
-               else             // error
-                  Tf (0, "--set-env: error reading: %s", arg);
+      if (args->env_deltas[i].action == ENV_SET) {  // --set-env
+         if (strchr(arg, '=') != NULL) {
+            // argument is variable name & value
+            envs_set((char *[]){ arg }, 1, NULL, args->c.env_expand);
+         } else {
+            // argument is filename
+            char **lines = NULL;
+            int line_ct = 0;
+            FILE *fp;
+            Tf (fp = fopen(arg, "r"), "--set-env: can't open: %s", arg);
+            for (line_ct = 0; true; line_ct++) {
+               char *line;
+               size_t line_len = 0;  // don't care but getline(3) must write
+               errno = 0;
+               if (-1 == getline(&line, &line_len, fp)) {
+                  if (errno == 0)  // EOF
+                     break;        // note: line_ct not incremented
+                  else
+                     Tf (0, "--set-env: can't read: %s", arg);
+               }
+               if (line[strlen(line) - 1] == '\n')  // rm newline if present
+                  line[strlen(line) - 1] = 0;
+               T_ (lines = realloc(lines, (line_ct + 1) * sizeof(char *)));
+               lines[line_ct] = line;
             }
-            if (strlen(line) == 0 || line[0] == '\n')
-               continue;                    // skip empty line
-            if (line[strlen(line) - 1] == '\n')
-               line[strlen(line) - 1] = 0;  // remove newline
-            split(&name, &new_value, line, '=');
-            Te (name != NULL, "--set-env: no delimiter: %s:%d", arg, j);
-            Te (strlen(name) != 0, "--set-env: empty name: %s:%d", arg, j);
-            if (   strlen(new_value) >= 2
-                && new_value[0] == '\''
-                && new_value[strlen(new_value) - 1] == '\'') {
-               new_value[strlen(new_value) - 1] = 0;  // strip trailing quote
-               new_value++;                           // strip leading
-            }
-            INFO("environment: %s=%s", name, new_value);
-            Z_ (setenv(name, new_value, 1));
+            Zf (fclose(fp), "--set-env: can't close: %s", arg);
+            envs_set(lines, line_ct, arg, args->c.env_expand);
          }
-         fclose(fp);
-      } else {
-         T_ (args->env_deltas[i].action == UNSET_GLOB);
+      } else {  // --unset-env
+         T_ (args->env_deltas[i].action == ENV_UNSET);
          /* Removing variables from the environment is tricky, because there
             is no standard library function to iterate through the
             environment, and the environ global array can be re-ordered after
@@ -374,6 +435,9 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
    int i;
 
    switch (key) {
+   case -10: // --env-no-expand
+      args->c.env_expand = false;
+      break;
    case -2: // --private-home
       args->c.private_home = true;
       break;
@@ -389,11 +453,11 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       args->c.join_pid = parse_int(arg, false, "--join-pid");
       break;
    case -6: // --set-env
-      env_delta_append(&(args->env_deltas), SET_FILE, arg);
+      env_delta_append(&(args->env_deltas), ENV_SET, arg);
       break;
    case -7: // --unset-env
       Te (strlen(arg) > 0, "--unset-env: GLOB must have non-zero length");
-      env_delta_append(&(args->env_deltas), UNSET_GLOB, arg);
+      env_delta_append(&(args->env_deltas), ENV_UNSET, arg);
       break;;
    case -8: // --ch-ssh
       args->c.ch_ssh = true;
@@ -404,19 +468,26 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
    case 'c':
       args->initial_dir = arg;
       break;
-   case 'b':
-      for (i = 0; args->c.binds[i].src != NULL; i++) // count existing binds
-         ;
-      T_ (args->c.binds = realloc(args->c.binds, (i+2) * sizeof(struct bind)));
-      args->c.binds[i+1].src = NULL;                 // terminating zero
-      args->c.binds[i].src = strsep(&arg, ":");
-      T_ (args->c.binds[i].src != NULL);
-      if (arg)
-         args->c.binds[i].dst = arg;
-      else // arg is NULL => no destination specified
-         T_ (1 <= asprintf(&(args->c.binds[i].dst), "/mnt/%d", i));
-      Te (args->c.binds[i].src[0] != 0, "--bind: no source provided");
-      Te (args->c.binds[i].dst[0] != 0, "--bind: no destination provided");
+   case 'b': {
+         char *src, *dst;
+         for (i = 0; args->c.binds[i].src != NULL; i++) // count existing binds
+            ;
+         T_ (args->c.binds = realloc(args->c.binds,
+                                     (i+2) * sizeof(struct bind)));
+         args->c.binds[i+1].src = NULL;                 // terminating zero
+         args->c.binds[i].dep = BD_MAKE_DST;
+         // source
+         src = strsep(&arg, ":");
+         T_ (src != NULL);
+         Te (src[0] != 0, "--bind: no source provided");
+         args->c.binds[i].src = src;
+         // destination
+         dst = arg ? arg : src;
+         Te (dst[0] != 0, "--bind: no destination provided");
+         Te (strcmp(dst, "/"), "--bind: destination can't be /");
+         Te (dst[0] == '/', "--bind: destination must be absolute");
+         args->c.binds[i].dst = dst;
+      }
       break;
    case 'g':
       i = parse_int(arg, false, "--gid");
@@ -440,7 +511,7 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       break;
    case 'v':
       verbose++;
-      Te(verbose <= 3, "--verbose can be specified at most twice");
+      Te(verbose <= 4, "--verbose can be specified at most thrice");
       break;
    case 'w':
       args->c.writable = true;
