@@ -22,7 +22,7 @@ builder_choose () {
         fi
     fi
     case $CH_BUILDER in
-        buildah|buildah-runc|buildah-setuid|ch-image|ch-grow|docker|none)
+        buildah|buildah-runc|buildah-setuid|ch-image|docker|none)
             ;;
         *)
             echo "unknown builder: $CH_BUILDER" 1>&2
@@ -31,6 +31,12 @@ builder_choose () {
     esac
 }
 
+# Return success if path $1 exists, without dereferencing links, failure
+# otherwise. ("test -e" dereferences.)
+exist_p () {
+    stat "$1" > /dev/null 2>&1
+}
+
 pack_fmt_valid () {
     case $1 in
     squash)
@@ -82,8 +88,11 @@ parse_basic_args () {
 }
 
 # Convert container registry path to filesystem compatible path.
+#
+# NOTE: This is used both to name user-visible stuff like tarballs as well as
+# dig around in the ch-image storage directory.
 tag_to_path () {
-    echo "$1" | sed 's/\//./g'
+    echo "$1" | sed 's/\//%/g'
 }
 
 usage () {
