@@ -3,16 +3,15 @@
 import abc
 import ast
 import glob
-import inspect
 import os
 import os.path
-import pathlib
 import re
 import shutil
 import sys
 
 import charliecloud as ch
 import fakeroot
+import pull
 
 
 ## Globals ##
@@ -54,13 +53,12 @@ ARG_DEFAULTS = { "HTTP_PROXY": os.environ.get("HTTP_PROXY"),
                  "https_proxy": os.environ.get("https_proxy"),
                  "ftp_proxy": os.environ.get("ftp_proxy"),
                  "no_proxy": os.environ.get("no_proxy"),
+                 "SSH_AUTH_SOCK": os.environ.get("SSH_AUTH_SOCK"),
                  "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
                  # GNU tar, when it thinks it's running as root, tries to
                  # chown(2) and chgrp(2) files to whatever's in the tarball.
                  "TAR_OPTIONS": "--no-same-owner" }
 
-ENV_DEFAULTS = { }
-
 
 ## Main ##
 
@@ -80,13 +78,28 @@ def main(cli_):
 
    # Infer image name if needed.
    if (cli.tag is None):
-      m = re.search(r"(([^/]+)/)?Dockerfile(\.(.+))?$",
-                    os.path.abspath(cli.file))
-      if (m is not None):
-         if m.group(4):    # extension
-            cli.tag = m.group(4)
-         elif m.group(2):  # containing directory
-            cli.tag = m.group(2)
+      path = os.path.basename(cli.file)
+      if ("." in path):
+         (base, ext_all) = str(path).split(".", maxsplit=1)
+         (base_all, ext_last) = str(path).rsplit(".", maxsplit=1)
+      else:
+         base = None
+         ext_last = None
+      if (base == "Dockerfile"):
+         cli.tag = ext_all
+         ch.VERBOSE("inferring name from Dockerfile extension: %s" % cli.tag)
+      elif (ext_last == "dockerfile"):
+         cli.tag = base_all
+         ch.VERBOSE("inferring name from Dockerfile basename: %s" % cli.tag)
+      elif (os.path.abspath(cli.context) != "/"):
+         cli.tag = os.path.basename(os.path.abspath(cli.context))
+         ch.VERBOSE("inferring name from context directory: %s" % cli.tag)
+      else:
+         assert (os.path.abspath(cli.context) == "/")
+         cli.tag = "root"
+         ch.VERBOSE("inferring name with root context directory: %s" % cli.tag)
+      cli.tag = re.sub(r"[^a-z0-9_.-]", "", cli.tag.lower())
+      ch.INFO("inferred image name: %s" % cli.tag)
 
    # Deal with build arguments.
    def build_arg_get(arg):
@@ -99,10 +112,7 @@ def main(cli_):
             ch.FATAL("--build-arg: %s: no value and not in environment" % kv[0])
          return (kv[0], v)
    cli.build_arg = dict( build_arg_get(i) for i in cli.build_arg )
-
-   # Finish CLI initialization.
    ch.DEBUG(cli)
-   ch.dependencies_check()
 
    # Guess whether the context is a URL, and error out if so. This can be a
    # typical looking URL e.g. "https://..." or also something like
@@ -137,9 +147,10 @@ def main(cli_):
    try:
       tree = parser.parse(text)
    except lark.exceptions.UnexpectedInput as x:
-      ch.DEBUG(x)  # noise about what was expected in the grammar
-      ch.FATAL("can't parse: %s:%d,%d\n\n%s" % (cli.file, x.line, x.column, x.get_context(text, 39)))
-   ch.DEBUG(tree.pretty())
+      ch.VERBOSE(x)  # noise about what was expected in the grammar
+      ch.FATAL("can't parse: %s:%d,%d\n\n%s"
+               % (cli.file, x.line, x.column, x.get_context(text, 39)))
+   ch.VERBOSE(tree.pretty())
 
    # Sometimes we exit after parsing.
    if (cli.parse_only):
@@ -208,6 +219,8 @@ class Main_Loop(lark.Visitor):
             else:
                ch.FATAL("first instruction must be ARG or FROM")
          inst.execute()
+         if (image_i != -1):
+            images[image_i].metadata_save()
          self.instruction_ct += inst.execute_increment
 
 
@@ -357,11 +370,17 @@ class I_copy(Instruction):
          paths = [variables_sub(i, env.env_build)
                   for i in ch.tree_child_terminals(self.tree, "copy_shell",
                                                    "WORD")]
-         self.srcs = paths[:-1]
-         self.dst = paths[-1]
+      elif (ch.tree_child(self.tree, "copy_list") is not None):
+         paths = [variables_sub(i, env.env_build)
+                  for i in ch.tree_child_terminals(self.tree, "copy_list",
+                                                   "STRING_QUOTED")]
+         for i in range(len(paths)):
+            paths[i] = paths[i][1:-1] #strip quotes
       else:
-         assert (ch.tree_child(self.tree, "copy_list") is not None)
-         self.unsupported_yet_fatal("list form", 784)
+         assert False, "unreachable code reached"
+
+      self.srcs = paths[:-1]
+      self.dst = paths[-1]
 
    def str_(self):
       return "%s -> %s" % (self.srcs, repr(self.dst))
@@ -374,43 +393,43 @@ class I_copy(Instruction):
          must exist already and be a directory. Unlike subdirectories, the
          metadata of dst will not be altered to match src."""
       def onerror(x):
-         ch.FATAL("error scanning directory: %s: %s" % (x.filename, x.strerror))
+         ch.FATAL("can't scan directory: %s: %s" % (x.filename, x.strerror))
       # Use Path objects in this method because the path arithmetic was
       # getting too hard with strings.
-      src = pathlib.Path(os.path.realpath(src))
-      dst = pathlib.Path(dst)
+      src = ch.Path(os.path.realpath(src))
+      dst = ch.Path(dst)
       assert (os.path.isdir(src) and not os.path.islink(src))
       assert (os.path.isdir(dst) and not os.path.islink(dst))
-      ch.DEBUG("copying named directory: %s -> %s" % (src, dst), v=2)
+      ch.DEBUG("copying named directory: %s -> %s" % (src, dst))
       for (dirpath, dirnames, filenames) in os.walk(src, onerror=onerror):
-         dirpath = pathlib.Path(dirpath)
+         dirpath = ch.Path(dirpath)
          subdir = dirpath.relative_to(src)
-         dst_dir = dst / subdir
+         dst_dir = dst // subdir
          # dirnames can contain symlinks, which we handle as files, so we'll
          # rebuild it; the walk will not descend into those "directories".
          dirnames2 = dirnames.copy()  # shallow copy
          dirnames[:] = list()         # clear in place
          for d in dirnames2:
-            d = pathlib.Path(d)
-            src_path = dirpath / d
-            dst_path = dst_dir / d
-            ch.DEBUG("dir: %s -> %s" % (src_path, dst_path), v=2)
+            d = ch.Path(d)
+            src_path = dirpath // d
+            dst_path = dst_dir // d
+            ch.TRACE("dir: %s -> %s" % (src_path, dst_path))
             if (os.path.islink(src_path)):
                filenames.append(d)  # symlink, handle as file
-               ch.DEBUG("symlink to dir, will handle as file", v=2)
+               ch.TRACE("symlink to dir, will handle as file")
                continue
             else:
                dirnames.append(d)   # directory, descend into later
             # If destination exists, but isn't a directory, remove it.
             if (os.path.exists(dst_path)):
                if (os.path.isdir(dst_path) and not os.path.islink(dst_path)):
-                  ch.DEBUG("dst_path exists and is a directory", v=2)
+                  ch.TRACE("dst_path exists and is a directory")
                else:
-                  ch.DEBUG("dst_path exists, not a directory, removing", v=2)
+                  ch.TRACE("dst_path exists, not a directory, removing")
                   ch.unlink(dst_path)
             # If destination directory doesn't exist, create it.
             if (not os.path.exists(dst_path)):
-               ch.DEBUG("mkdir dst_path", v=2)
+               ch.TRACE("mkdir dst_path")
                ch.ossafe(os.mkdir, "can't mkdir: %s" % dst_path, dst_path)
             # Copy metadata, now that we know the destination exists and is a
             # directory.
@@ -418,15 +437,15 @@ class I_copy(Instruction):
                       "can't copy metadata: %s -> %s" % (src_path, dst_path),
                       src_path, dst_path, follow_symlinks=False)
          for f in filenames:
-            f = pathlib.Path(f)
-            src_path = dirpath / f
-            dst_path = dst_dir / f
-            ch.DEBUG("file or symlink via copy2: %s -> %s"
-                     % (src_path, dst_path), v=2)
+            f = ch.Path(f)
+            src_path = dirpath // f
+            dst_path = dst_dir // f
+            ch.TRACE("file or symlink via copy2: %s -> %s"
+                      % (src_path, dst_path))
             if (not (os.path.isfile(src_path) or os.path.islink(src_path))):
                ch.FATAL("can't COPY: unknown file type: %s" % src_path)
             if (os.path.exists(dst_path)):
-               ch.DEBUG("destination exists, removing", v=2)
+               ch.TRACE("destination exists, removing")
                if (os.path.isdir(dst_path) and not os.path.islink(dst_path)):
                   ch.rmtree(dst_path)
                else:
@@ -444,7 +463,7 @@ class I_copy(Instruction):
       assert (   not os.path.exists(dst)
               or (os.path.isdir(dst) and not os.path.islink(dst))
               or (os.path.isfile(dst) and not os.path.islink(dst)))
-      ch.DEBUG("copying named file: %s -> %s" % (src, dst), v=2)
+      ch.DEBUG("copying named file: %s -> %s" % (src, dst))
       ch.copy2(src, dst, follow_symlinks=True)
 
    def dest_realpath(self, unpack_path, dst):
@@ -452,39 +471,41 @@ class I_copy(Instruction):
         path unpack_path. We can't use os.path.realpath() because if dst is
         an absolute symlink, we need to use the *image's* root directory, not
         the host. Thus, we have to resolve symlinks manually."""
-      unpack_path = pathlib.Path(unpack_path)
-      dst_canon = pathlib.Path(unpack_path)
-      dst = pathlib.Path(dst)
+      unpack_path = ch.Path(unpack_path)
+      dst_canon = ch.Path(unpack_path)
+      dst = ch.Path(dst)
       dst_parts = list(reversed(dst.parts))  # easier to operate on end of list
       iter_ct = 0
       while (len(dst_parts) > 0):
          iter_ct += 1
          if (iter_ct > 100):  # arbitrary
             ch.FATAL("can't COPY: too many path components")
-         ch.DEBUG("current destination: %d %s" % (iter_ct, dst_canon), v=2)
-         #ch.DEBUG("parts remaining: %s" % dst_parts, v=2)
+         ch.TRACE("current destination: %d %s" % (iter_ct, dst_canon))
+         #ch.TRACE("parts remaining: %s" % dst_parts)
          part = dst_parts.pop()
          if (part == "/" or part == "//"):  # 3 or more slashes yields "/"
-            ch.DEBUG("skipping root")
+            ch.TRACE("skipping root")
             continue
-         cand = dst_canon / part
-         ch.DEBUG("checking: %s" % cand, v=2)
+         cand = dst_canon // part
+         ch.TRACE("checking: %s" % cand)
          if (not cand.is_symlink()):
-            ch.DEBUG("not symlink", v=2)
+            ch.TRACE("not symlink")
             dst_canon = cand
          else:
-            target = pathlib.Path(os.readlink(cand))
-            ch.DEBUG("symlink to: %s" % target, v=2)
+            target = ch.Path(os.readlink(cand))
+            ch.TRACE("symlink to: %s" % target)
             assert (len(target.parts) > 0)  # POSIX says no empty symlinks
             if (target.is_absolute()):
-               ch.DEBUG("absolute")
-               dst_canon = pathlib.Path(unpack_path)
+               ch.TRACE("absolute")
+               dst_canon = ch.Path(unpack_path)
             else:
-               ch.DEBUG("relative", v=2)
+               ch.TRACE("relative")
             dst_parts.extend(reversed(target.parts))
       return dst_canon
 
    def execute_(self):
+      if (len(self.srcs) < 1):
+         ch.FATAL("can't COPY: must specify at least one source")
       # Complain about unsupported stuff.
       if (self.options.pop("chown", False)):
          self.unsupported_forever_warn("--chown")
@@ -509,15 +530,16 @@ class I_copy(Instruction):
                ch.FATAL("COPY --from: stage %s does not exist" % self.from_)
          context = images[self.from_].unpack_path
       context_canon = os.path.realpath(context)
-      ch.DEBUG("context: %s" % context)
+      ch.VERBOSE("context: %s" % context)
       # Expand source wildcards.
       srcs = list()
       for src in self.srcs:
-         for i in glob.glob(context + "/" + src):
+         matches = glob.glob("%s/%s" % (context, src))  # glob can't take Path
+         if (len(matches) == 0):
+            ch.FATAL("can't copy: not found: %s" % src)
+         for i in matches:
             srcs.append(i)
-            ch.DEBUG("source: %s" % i)
-      if (len(srcs) == 0):
-         ch.FATAL("can't COPY: no sources found")
+            ch.VERBOSE("source: %s" % i)
       # Validate sources are within context directory. (Can't convert to
       # canonical paths yet because we need the source path as given.)
       for src in srcs:
@@ -528,17 +550,17 @@ class I_copy(Instruction):
       # Locate the destination.
       unpack_canon = os.path.realpath(images[image_i].unpack_path)
       if (self.dst.startswith("/")):
-         dst = self.dst
+         dst = ch.Path(self.dst)
       else:
-         dst = env.workdir + "/" + self.dst
-      ch.DEBUG("destination, as given: %s" % dst)
+         dst = env.workdir // self.dst
+      ch.VERBOSE("destination, as given: %s" % dst)
       dst_canon = self.dest_realpath(unpack_canon, dst) # strips trailing slash
-      ch.DEBUG("destination, canonical: %s" % dst_canon)
+      ch.VERBOSE("destination, canonical: %s" % dst_canon)
       if (not os.path.commonpath([dst_canon, unpack_canon])
               .startswith(unpack_canon)):
          ch.FATAL("can't COPY: destination not in image: %s" % dst_canon)
       # Create the destination directory if needed.
-      if (dst.endswith("/") or len(srcs) > 1 or os.path.isdir(srcs[0])):
+      if (self.dst.endswith("/") or len(srcs) > 1 or os.path.isdir(srcs[0])):
          if (not os.path.exists(dst_canon)):
             ch.mkdirs(dst_canon)
          elif (not os.path.isdir(dst_canon)):  # not symlink b/c realpath()
@@ -569,7 +591,7 @@ class Env(Instruction):
 
    def execute_(self):
       env.env[self.key] = self.value
-      with ch.open_(images[image_i].unpack_path + "/ch/environment", "wt") \
+      with ch.open_(images[image_i].unpack_path // "/ch/environment", "wt") \
            as fp:
          for (k, v) in env.env.items():
             print("%s=%s" % (k, v), file=fp)
@@ -580,10 +602,10 @@ class I_env_equals(Env):
    def __init__(self, *args):
       super().__init__(*args)
       self.key = ch.tree_terminal(self.tree, "WORD", 0)
-      self.value = ch.tree_terminal(self.tree, "WORD", 1)
-      if (self.value is None):
-         self.value = unescape(ch.tree_terminal(self.tree, "STRING_QUOTED"))
-      self.value = variables_sub(self.value, env.env_build)
+      v = ch.tree_terminal(self.tree, "WORD", 1)
+      if (v is None):
+         v = ch.tree_terminal(self.tree, "STRING_QUOTED")
+      self.value = variables_sub(unescape(v), env.env_build)
 
 
 class I_env_space(Env):
@@ -591,11 +613,8 @@ class I_env_space(Env):
    def __init__(self, *args):
       super().__init__(*args)
       self.key = ch.tree_terminal(self.tree, "WORD")
-      value = ch.tree_terminal(self.tree, "LINE")
-      if (not value.startswith('"')):
-         value = '"' + value + '"'
-      self.value = unescape(value)
-      self.value = variables_sub(self.value, env.env_build)
+      v = ch.tree_terminals_cat(self.tree, "LINE_CHUNK")
+      self.value = variables_sub(unescape(v), env.env_build)
 
 
 class I_from_(Instruction):
@@ -627,22 +646,26 @@ class I_from_(Instruction):
       else:
          # Not last image; append stage index to tag.
          tag = "%s/_stage%d" % (cli.tag, image_i)
-      image = ch.Image(ch.Image_Ref(tag), cli.storage + "/dlcache",
-                       cli.storage + "/img")
+      image = ch.Image(ch.Image_Ref(tag))
       images[image_i] = image
       if (self.alias is not None):
          images[self.alias] = image
-      ch.DEBUG("image path: %s" % image.unpack_path)
+      ch.VERBOSE("image path: %s" % image.unpack_path)
       # Other error checking.
       if (str(image.ref) == str(self.base_ref)):
          ch.FATAL("output image ref same as FROM: %s" % self.base_ref)
       # Initialize image.
-      self.base_image = ch.Image(self.base_ref, image.download_cache,
-                                 image.unpack_dir)
-      if (not os.path.isdir(self.base_image.unpack_path)):
-         ch.DEBUG("image not found, pulling: %s" % self.base_image.unpack_path)
-         self.base_image.pull_to_unpacked(fixup=True)
+      self.base_image = ch.Image(self.base_ref)
+      if (os.path.isdir(self.base_image.unpack_path)):
+         ch.VERBOSE("base image found: %s" % self.base_image.unpack_path)
+      else:
+         ch.VERBOSE("base image not found, pulling")
+         # a young hen, especially one less than one year old.
+         pullet = pull.Image_Puller(self.base_image, not cli.no_cache)
+         pullet.pull_to_unpacked()
+         pullet.done()
       image.copy_unpacked(self.base_image)
+      image.metadata_load()
       env.reset()
       # Find fakeroot configuration, if any.
       global fakeroot_config
@@ -650,8 +673,8 @@ class I_from_(Instruction):
                                         cli.force, cli.no_force_detect)
 
    def str_(self):
-      alias = "AS %s" % self.alias if self.alias else ""
-      return "%s %s" % (self.base_ref, alias)
+      alias = " AS %s" % self.alias if self.alias else ""
+      return "%s%s" % (self.base_ref, alias)
 
 
 class Run(Instruction):
@@ -689,18 +712,33 @@ class I_run_exec(Run):
 
 class I_run_shell(Run):
 
+   # Note re. line continuations and whitespace: Whitespace before the
+   # backslash is passed verbatim to the shell, while the newline and any
+   # whitespace between the newline and baskslash are deleted.
+
    def __init__(self, *args):
       super().__init__(*args)
-      # FIXME: Can't figure out how to remove continuations at parse time.
-      cmd = ch.tree_terminal(self.tree, "LINE").replace("\\\n", "")
-      self.cmd = ["/bin/sh", "-c", cmd]
+      cmd = ch.tree_terminals_cat(self.tree, "LINE_CHUNK")
+      self.cmd = env.shell + [cmd]
 
+class I_shell(Instruction):
+
+   def __init__(self, *args):
+      super().__init__(*args)
+      self.shell = [variables_sub(unescape(i), env.env_build)
+                    for i in ch.tree_terminals(self.tree, "STRING_QUOTED")]
+
+   def str_(self):
+      return str(self.shell)
+
+   def execute_(self):
+      env.shell = list(self.shell) #copy
 
 class I_workdir(Instruction):
 
    def __init__(self, *args):
       super().__init__(*args)
-      self.path = variables_sub(ch.tree_terminal(self.tree, "LINE"),
+      self.path = variables_sub(ch.tree_terminals_cat(self.tree, "LINE_CHUNK"),
                                 env.env_build)
 
    def str_(self):
@@ -708,7 +746,7 @@ class I_workdir(Instruction):
 
    def execute_(self):
       env.chdir(self.path)
-      ch.mkdirs(images[image_i].unpack_path + env.workdir)
+      ch.mkdirs(images[image_i].unpack_path // env.workdir)
 
 
 class I_uns_forever(Instruction_Supported_Never):
@@ -732,8 +770,7 @@ class I_uns_yet(Instruction):
                         "CMD":         780,
                         "ENTRYPOINT":  780,
                         "LABEL":       781,
-                        "ONBUILD":     788,
-                        "SHELL":       789 }[self.name]
+                        "ONBUILD":     788 }[self.name]
 
    def announce(self):
       self.unsupported_yet_warn("instruction", self.issue_no)
@@ -751,25 +788,57 @@ class I_uns_yet(Instruction):
 ## Supporting classes ##
 
 class Environment:
-   "The state we are in: environment variables, working directory, etc."
+   """The state we are in: environment variables, working directory, etc. Most
+      of this is just passed through from the image metadata."""
+
+   __slots__ = ("arg",)
 
    def __init__(self):
       self.reset()
 
+   @property
+   def env(self):
+      if (image_i == -1):
+         return dict()
+      else:
+         return images[image_i].metadata["env"]
+
+   @env.setter
+   def env(self, x):
+      images[image_i].metadata["env"] = x
+
    @property
    def env_build(self):
       return { **self.arg, **self.env }
 
+   @property
+   def shell(self):
+      if (image_i == -1):
+         return ["/bin/false"]
+      else:
+         return images[image_i].metadata["shell"]
+
+   @shell.setter
+   def shell(self, x):
+      images[image_i].metadata["shell"] = x
+
+   @property
+   def workdir(self):
+      return ch.Path(images[image_i].metadata["cwd"])
+
+   @workdir.setter
+   def workdir(self, x):
+      images[image_i].metadata["cwd"] = str(x)
+
    def chdir(self, path):
       if (path.startswith("/")):
-         self.workdir = path
+         self.workdir = ch.Path(path)
       else:
-         self.workdir += "/" + path
+         self.workdir //= path
 
    def reset(self):
-      self.workdir = "/"
+      # This resets only things that aren't part of the image metadata.
       self.arg = { k: v for (k, v) in ARG_DEFAULTS.items() if v is not None }
-      self.env = { k: v for (k, v) in ENV_DEFAULTS.items() if v is not None }
 
 
 ## Supporting functions ###
@@ -795,6 +864,8 @@ def unescape(sl):
    # guessing it's the Go rules. You will note that we are using Python rules.
    # This is wrong but close enough for now (see also gripe in previous
    # paragraph).
-   if (not (sl.startswith('"') and sl.endswith('"'))):
-      ch.FATAL("string literal not quoted")
+   if (    not sl.startswith('"')                          # no start quote
+       and (not sl.endswith('"') or sl.endswith('\\"'))):  # no end quote
+      sl = '"%s"' % sl
+   assert (len(sl) >= 2 and sl[0] == '"' and sl[-1] == '"' and sl[-2:] != '\\"')
    return ast.literal_eval(sl)
