@@ -135,7 +135,7 @@ DEFAULT_CONFIGS = {
    #
    # 5. Why grep a specified file vs. simpler alternatives?
    #
-   #    * Look at image name: Misses derived images, large number of tags
+   #    * Look at image name: Misses derived images, large number of names
    #      seems a maintenance headache, :latest changes.
    #
    #    * grep the same file for each distro: No standardized file for this.
@@ -147,17 +147,17 @@ DEFAULT_CONFIGS = {
    # 1. CentOS seems to have only fakeroot, which is in EPEL, not the standard
    #    repos.
    #
-   # 2. Enabling EPEL can have undesirable side effects, e.g. different
+   # 2. Unlike on CentOS, RHEL doesn't have the epel-release rpm in the
+   #    standard repos, install via rpm for both to be consistent.
+   #
+   # 3. Enabling EPEL can have undesirable side effects, e.g. different
    #    version of things in the base repo that breaks other things. Thus,
-   #    when we install EPEL, we don't enable it. Existing EPEL installations
-   #    are left alone.
+   #    when we are done with EPEL, we uninstall it. Existing EPEL
+   #    installations are left alone.
    #
-   # 3. "yum repolist" has a lot of side effects, e.g. locking the RPM
+   # 4. "yum repolist" has a lot of side effects, e.g. locking the RPM
    #    database and asking configured repos for something or other.
-   #
-   # 4. "dnf config-manager" (CentOS 8) requires installing dnf-plugins-core,
-   #    which requires fakeroot, which we don't have when initializing
-   #    fakeroot. So sed it is. :P
+
 
    "rhel7":
    { "name": "CentOS/RHEL 7",
@@ -165,34 +165,38 @@ DEFAULT_CONFIGS = {
      "init": [ ("command -v fakeroot > /dev/null",
                 "set -ex; "
                 "if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
-                "yum install -y epel-release; "
-                "yum-config-manager --disable epel; "
-                "fi; "
-                "yum --enablerepo=epel install -y fakeroot; ") ],
+                "yum install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm; "
+                "yum install -y fakeroot; "
+                "yum remove -y epel-release; "
+                "else "
+                "yum install -y fakeroot; "
+                "fi; ") ],
      "cmds": ["dnf", "rpm", "yum"],
      "each": ["fakeroot"] },
 
    "rhel8":
    { "name": "CentOS/RHEL 8",
-     "match":  ("/etc/redhat-release", r"release 8\."),
+     "match":  ("/etc/redhat-release", r"release 8"),
      "init": [ ("command -v fakeroot > /dev/null",
                 "set -ex; "
                 "if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
-                "dnf install -y epel-release; "
+                "dnf install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm; "
                 "ls -lh /etc/yum.repos.d; "
-                "sed -Ei 's/enabled=1$/enabled=0/g' /etc/yum.repos.d/epel*.repo; "
-                "fi; "
-                "dnf --enablerepo=epel install -y fakeroot; ") ],
+                "dnf install -y fakeroot; "
+                "dnf remove -y epel-release; "
+                "else "
+                "dnf install -y fakeroot; "
+                "fi; ") ],
      "cmds": ["dnf", "rpm", "yum"],
      "each": ["fakeroot"] },
 
-   # Debian notes:
+   # Debian/Ubuntu notes:
    #
-   # 1. By default in recent Debians, apt(8) runs as an unprivileged user.
-   #    This makes *all* apt operations fail in an unprivileged container
-   #    because it can't drop privileges. There are multiple ways to turn the
-   #    “sandbox” off. As far as I can tell, none are documented, but this one
-   #    at least appears in google searches a lot.
+   # 1. In recent Debian-based distributions apt(8) runs as an unprivileged
+   #    user by default. This makes *all* apt operations fail in an
+   #    unprivileged container because it can't drop privileges. There are
+   #    multiple ways to turn the “sandbox” off. AFAICT, none are documented,
+   #    but this one at least appears in Google searches a lot.
    #
    #    apt also doesn't drop privileges if there is no user _apt; in my
    #    testing, sometimes this user is present and sometimes not, for reasons
@@ -208,9 +212,9 @@ DEFAULT_CONFIGS = {
    #      dpkg-query -Wf '${Package}\n' \
    #      | egrep '^(fakeroot|fakeroot-ng|pseudo)$'
 
-   "debSB":
-   { "name": "Debian 9 (Stretch) or 10 (Buster)",
-     "match": ("/etc/debian_version", r"^(9|10)\."),
+   "debderiv":
+   { "name": "Debian (9, 10, 11) or Ubuntu (16, 18, 20)",
+     "match": ("/etc/os-release", r"(stretch|buster|bullseye|xenial|bionic|focal)"),
      "init": [ ("apt-config dump | fgrep -q 'APT::Sandbox::User \"root\"'"
                 " || ! fgrep -q _apt /etc/passwd",
                 "echo 'APT::Sandbox::User \"root\";'"
@@ -221,6 +225,22 @@ DEFAULT_CONFIGS = {
      "cmds": ["apt", "apt-get", "dpkg"],
      "each": ["fakeroot"] },
 
+   # Fedora notes:
+   #
+   # 1. The minimum supported version was chosen somewhat arbitrarily based on the
+   #    release versions available for testing (what was on Docker Hub).
+   #
+   # 2. The fakeroot package is in the base repository set so enabling EPEL is
+   #    not required.
+
+   "fedora":
+   { "name": "Fedora 24+,",
+     "match":  ("/etc/fedora-release", r"release (2[4-9]|[3-9][0-9])"),
+     "init": [ ("command -v fakeroot > /dev/null",
+                "dnf install -y fakeroot") ],
+     "cmds": ["dnf", "rpm", "yum"],
+     "each": ["fakeroot"] },
+
 }
 
 
@@ -229,7 +249,7 @@ DEFAULT_CONFIGS = {
 def detect(image, force, no_force_detect):
    f = None
    if (no_force_detect):
-      ch.DEBUG("not detecting --force config, per --no-force-detect")
+      ch.VERBOSE("not detecting --force config, per --no-force-detect")
    else:
       # Try to find a real fakeroot config.
       for (tag, cfg) in DEFAULT_CONFIGS.items():
@@ -244,7 +264,7 @@ def detect(image, force, no_force_detect):
          if (force):
             ch.WARNING(msg)
          else:
-            ch.DEBUG(msg)
+            ch.VERBOSE(msg)
       else:
          if (force):
             adj = "will use"
@@ -289,7 +309,7 @@ class Fakeroot():
                 "inject_p")
 
    def __init__(self, image_path, tag, cfg, inject_p):
-      ch.DEBUG("workarounds: testing config: %s" % tag)
+      ch.VERBOSE("workarounds: testing config: %s" % tag)
       file_path = "%s/%s" % (image_path, cfg["match"][0])
       if (not (    os.path.isfile(file_path)
                and ch.grep_p(file_path, cfg["match"][1]))):
@@ -303,10 +323,10 @@ class Fakeroot():
 
    def init_maybe(self, img_path, args, env):
       if (not self.needs_inject(args)):
-         ch.DEBUG("workarounds: init: instruction doesn't need injection")
+         ch.VERBOSE("workarounds: init: instruction doesn't need injection")
          return
       if (self.init_done):
-         ch.DEBUG("workarounds: init: already initialized")
+         ch.VERBOSE("workarounds: init: already initialized")
          return
       for (i, (test_cmd, init_cmd)) in enumerate(self.init, 1):
          ch.INFO("workarounds: init step %s: checking: $ %s" % (i, test_cmd))
@@ -326,7 +346,7 @@ class Fakeroot():
 
    def inject_run(self, args):
       if (not self.needs_inject(args)):
-         ch.DEBUG("workarounds: RUN: instruction doesn't need injection")
+         ch.VERBOSE("workarounds: RUN: instruction doesn't need injection")
          return args
       assert (self.init_done)
       if (not self.inject_p):
